part of gmo_camera_picker;

class CameraPickerViewer extends StatefulWidget {
  const CameraPickerViewer({
    Key? key,
    required this.pickerState,
    required this.pickerType,
    required this.previewXFile,
    required this.theme,
    this.shouldAutoPreviewVideo = false,
  }) : super(key: key);

  final GmoCameraPickerState pickerState;

  final CameraPickerViewType pickerType;

  final XFile previewXFile;

  final ThemeData theme;

  final bool shouldAutoPreviewVideo;

  static Future<File?> pushToViewer(
    BuildContext context, {
    required GmoCameraPickerState pickerState,
    required CameraPickerViewType pickerType,
    required XFile previewXFile,
    required ThemeData theme,
    bool shouldAutoPreviewVideo = false,
  }) {
    return Navigator.of(context).push<File?>(
      PageRouteBuilder<File?>(
        pageBuilder: (_, __, ___) => CameraPickerViewer(
          pickerState: pickerState,
          pickerType: pickerType,
          previewXFile: previewXFile,
          theme: theme,
          shouldAutoPreviewVideo: shouldAutoPreviewVideo,
        ),
        transitionsBuilder: (
          BuildContext context,
          Animation<double> animation,
          Animation<double> secondaryAnimation,
          Widget child,
        ) {
          return FadeTransition(opacity: animation, child: child);
        },
      ),
    );
  }

  @override
  _CameraPickerViewerState createState() => _CameraPickerViewerState();
}

class _CameraPickerViewerState extends State<CameraPickerViewer> {
  late final VideoPlayerController videoController =
      VideoPlayerController.file(previewFile);

  late bool hasLoaded = pickerType == CameraPickerViewType.image;

  bool hasErrorWhenInitializing = false;

  final ValueNotifier<bool> isPlaying = ValueNotifier<bool>(false);

  bool get isControllerPlaying => videoController.value.isPlaying;

  GmoCameraPickerState get pickerState => widget.pickerState;

  CameraPickerViewType get pickerType => widget.pickerType;

  XFile get previewXFile => widget.previewXFile;

  File get previewFile => File(previewXFile.path);

  ThemeData get theme => widget.theme;

  bool get shouldAutoPreviewVideo => widget.shouldAutoPreviewVideo;

  @override
  void initState() {
    super.initState();
    if (pickerType == CameraPickerViewType.video) {
      initializeVideoPlayerController();
    }
  }

  @override
  void dispose() {
    videoController.removeListener(videoPlayerListener);
    videoController.pause();
    videoController.dispose();
    super.dispose();
  }

  Future<void> initializeVideoPlayerController() async {
    try {
      await videoController.initialize();
      videoController.addListener(videoPlayerListener);
      hasLoaded = true;
      if (shouldAutoPreviewVideo) {
        videoController.play();
      }
    } catch (e) {
      hasErrorWhenInitializing = true;
      realDebugPrint('Error when initializing video controller: $e');
    } finally {
      if (mounted) {
        setState(() {});
      }
    }
  }

  void videoPlayerListener() {
    if (isControllerPlaying != isPlaying.value) {
      isPlaying.value = isControllerPlaying;
    }
  }

  Future<void> playButtonCallback() async {
    if (isPlaying.value) {
      videoController.pause();
    } else {
      if (videoController.value.duration == videoController.value.position) {
        videoController
          ..seekTo(Duration.zero)
          ..play();
      } else {
        videoController.play();
      }
    }
  }

  Future<void> createAssetEntityAndPop() async {
    Navigator.of(context).pop(previewFile);
  }

  Widget get previewBackButton {
    return InkWell(
      borderRadius: maxBorderRadius,
      onTap: () {
        if (previewFile.existsSync()) {
          previewFile.delete();
        }
        Navigator.of(context).pop();
      },
      child: Container(
        margin: const EdgeInsets.all(10.0),
        width: Screens.width / 15,
        height: Screens.width / 15,
        decoration: const BoxDecoration(
          color: Colors.white,
          shape: BoxShape.circle,
        ),
        child: const Center(
          child: Icon(
            Icons.keyboard_return_rounded,
            color: Colors.black,
            size: 18.0,
          ),
        ),
      ),
    );
  }

  Widget get previewConfirmButton {
    return MaterialButton(
      minWidth: 20.0,
      height: 32.0,
      padding: const EdgeInsets.symmetric(horizontal: 20.0),
      color: theme.colorScheme.secondary,
      shape: RoundedRectangleBorder(
        borderRadius: BorderRadius.circular(3.0),
      ),
      child: Text(
        Constants.textDelegate.confirm,
        style: TextStyle(
          color: theme.textTheme.bodyText1?.color,
          fontSize: 17.0,
          fontWeight: FontWeight.normal,
        ),
      ),
      onPressed: createAssetEntityAndPop,
      materialTapTargetSize: MaterialTapTargetSize.shrinkWrap,
    );
  }

  Widget get playControlButton {
    return ValueListenableBuilder<bool>(
      valueListenable: isPlaying,
      builder: (_, bool value, Widget? child) => GestureDetector(
        behavior: HitTestBehavior.opaque,
        onTap: value ? playButtonCallback : null,
        child: Center(
          child: AnimatedOpacity(
            duration: kThemeAnimationDuration,
            opacity: value ? 0.0 : 1.0,
            child: GestureDetector(
              onTap: playButtonCallback,
              child: DecoratedBox(
                decoration: const BoxDecoration(
                  boxShadow: <BoxShadow>[BoxShadow(color: Colors.black12)],
                  shape: BoxShape.circle,
                ),
                child: Icon(
                  value ? Icons.pause_circle_outline : Icons.play_circle_filled,
                  size: 70.0,
                  color: Colors.white,
                ),
              ),
            ),
          ),
        ),
      ),
    );
  }

  Widget get viewerActions {
    return SafeArea(
      child: Padding(
        padding: const EdgeInsets.symmetric(
          horizontal: 8.0,
          vertical: 20.0,
        ),
        child: Column(
          mainAxisAlignment: MainAxisAlignment.spaceBetween,
          children: <Widget>[
            Row(
              children: <Widget>[
                previewBackButton,
                const Spacer(),
              ],
            ),
            Row(
              children: <Widget>[
                const Spacer(),
                previewConfirmButton,
              ],
            ),
          ],
        ),
      ),
    );
  }

  @override
  Widget build(BuildContext context) {
    if (hasErrorWhenInitializing) {
      return Center(
        child: Text(
          Constants.textDelegate.loadFailed,
          style: const TextStyle(inherit: false),
        ),
      );
    }
    if (!hasLoaded) {
      return const SizedBox.shrink();
    }
    return Material(
      color: Colors.black,
      child: Stack(
        children: <Widget>[
          if (pickerType == CameraPickerViewType.image)
            Positioned.fill(child: Image.file(previewFile))
          else if (pickerType == CameraPickerViewType.video)
            Positioned.fill(
              child: Center(
                child: AspectRatio(
                  aspectRatio: videoController.value.aspectRatio,
                  child: VideoPlayer(videoController),
                ),
              ),
            ),
          if (pickerType == CameraPickerViewType.video) playControlButton,
          viewerActions,
        ],
      ),
    );
  }
}
